@@ -792,7 +792,45 @@ def get_pulsed_fraction(e_min, e_max, pp, dpp, method_calc_rms='adaptive', outpu
 
 def elaborate_pulsed_fraction(ee_pulsed, dee_pulsed, pulsed_frac, dpulsed_frac, debug_plots=True, e1=10, e2=20, 
                               stem='', poly_deg=3, title='', save_plot=True, e_threshold=0.3, 
-                              division_derivative_order=2):
+                              division_derivative_order=2, max_n_high_lines = 2, forced_gaussian_centroids = [],
+                              forced_gaussian_amplitudes = [],
+                              forced_gaussian_sigmas = []):
+    """This function implements the logics to perform a fit of the pulsed fraction.
+    It first interpolates the pulsed fraction with a spline to find extremes of the function.
+    Then it performs a fit with a polynomial plus gaussians.
+    The general idea is to split the pulsed fraction into two regions, based on the flex of the function.
+    Then to fit the lower energy checking for a gaussian feature in correspondence of the iron line complex.
+    Then we look for minima of the PF in the highe part and inet gussians. 
+    If there are too may minima, we reduce them.
+    Logic is included to comply with few points.
+
+    Args:
+        ee_pulsed (numpy array): energy
+        dee_pulsed (numpy array): energy error
+        pulsed_frac (numpy array): pulsed fraction
+        dpulsed_frac (numpy array): pulsed fraction uncertainty
+        debug_plots (bool, optional): plot the derivative and the fits. Defaults to True.
+        e1 (int, optional): lower energy to search the flex in this interval. Defaults to 10.
+        e2 (int, optional): lower energy to search the flex in this interval. Defaults to 20.
+        Note that if you put e1>=e2, we will use only one interval.
+        stem (str, optional): a string in fron of output plot names. Defaults to ''.
+        poly_deg (int, optional): degree of the polynomial. Defaults to 3.
+        title (str, optional): title of the plots. Defaults to ''.
+        save_plot (bool, optional): if plots should be saved Defaults to True.
+        e_threshold (float, optional): if energies of PF minima differ in relative terms less than this values, they are reduced.
+                                       ex.: if the algorithm finds minima at 32 and 35 keV, it retains only 32.
+                                       This is used also as a range for the centroid search from (1-e_treshold)*e_centroid to 
+                                       (1+e_treshold)*e_centroid Defaults to 0.3.
+        division_derivative_order (int, optional): order of the derivative to search for the division, if 2 it searches a flex. Defaults to 2.
+        max_n_high_lines (int, optional): do not use more gaussians thn this. Defaults to 2.
+        forced_gaussian_centroids (list, optional): you can force the initial centroid energies with this parameter. Defaults to [].
+        forced_gaussian_amplitudes (list, optional): you can force the initial amplitudes with this parameter, if the length of the list 
+                                                    is different from the one of forced_gaussian_centroids, it is unused. Defaults to [].
+        forced_gaussian_sigmas (list, optional): you can force the initial sigmas with this parameter, if the length of the list 
+                                                 is different from the one of forced_gaussian_centroids, it is unused. Defaults to [].
+    Returns:
+        two fitting objects for the two ranges (the output of utils.fit_pulsed_frac). if only one is used, the second is None.
+     """    
     from scipy import interpolate
     smoothing = interpolate.UnivariateSpline(ee_pulsed, pulsed_frac, w=1./dpulsed_frac, k=3)
     fake_en = np.linspace(ee_pulsed[0], ee_pulsed[-1], 1000)
@@ -803,10 +841,12 @@ def elaborate_pulsed_fraction(ee_pulsed, dee_pulsed, pulsed_frac, dpulsed_frac,
         positive = deriv[selection] > 0
         ind_extreme = np.where(np.bitwise_xor(positive[1:], positive[:-1]))
         #print(energy[selection][ind_extreme])
-        ind_zeros = deriv_2[selection][ind_extreme] > 0
-        #print(deriv_2[selection][ind_extreme])
-        #print(ind_zeros)
-        return energy[selection][ind_extreme][ind_zeros]
+        if deriv_2 is not None:
+            ind_zeros = deriv_2[selection][ind_extreme] > 0
+            return energy[selection][ind_extreme][ind_zeros]
+        else:
+            ind_zeros = ind_extreme
+            return energy[selection][ind_zeros]
 
     def get_minimum(energy, deriv, selection):
         ind_min = np.argmin(deriv[selection])
@@ -826,91 +866,130 @@ def elaborate_pulsed_fraction(ee_pulsed, dee_pulsed, pulsed_frac, dpulsed_frac,
         axes[1].set_ylabel('Pulsed Fraction Derivative')
 
     ind_range1 = (fake_en >= e1) & (fake_en <=e2)
+    make_double_fit = True
+    if np.sum(ind_range1) <= 0:
+        make_double_fit = False
 
-    if division_derivative_order == 2:
-        e_turn = get_zeros(fake_en, smoothed_pulsed_frac_deriv_2, None, ind_range1)
+
+    center=[[6.5, 5.5, 7.5]]
+    sigma=[[0.2,0,1.0]]
+    amplitude=[[0,-1,1]]
+    if make_double_fit:
+        if division_derivative_order == 2:
+            e_turn = get_zeros(fake_en, smoothed_pulsed_frac_deriv_2, None, ind_range1)
+        else:
+            e_turn = get_zeros(fake_en, smoothed_pulsed_frac_deriv, smoothed_pulsed_frac_deriv_2, ind_range1)
+        
+        if len(e_turn) == 0:
+            logger.info('Using the minimum derivative instead of zero in dividing regimes')
+            e_turn = get_minimum(fake_en, smoothed_pulsed_frac_deriv, ind_range1)
+        else:
+            logger.info('Getting a zero of derivative to divide regimes')
+            e_turn = e_turn[0]
+
+        logger.info("We separate PF fit at %f keV" % e_turn)
+
+        if debug_plots:
+            axes[0].axvline(e_turn, 0, 1, color='cyan')
+            axes[1].axvline(e_turn, 0, 1, color='cyan')
+
+        ind_low = ee_pulsed <= e_turn
+        ind_high = ee_pulsed > e_turn
+        logger.info('We use %d points in the low-energy' % np.sum(ind_low))
     else:
-        e_turn = get_zeros(fake_en, smoothed_pulsed_frac_deriv, smoothed_pulsed_frac_deriv_2, ind_range1)
-    
-    if len(e_turn) == 0:
-        logger.info('Using the minimum derivative instead of zero in dividing regimes')
-        e_turn = get_minimum(fake_en, smoothed_pulsed_frac_deriv, ind_range1)
-    else:
-        logger.info('Getting a zero of derivative to divide regimes')
-        e_turn = e_turn[0]
+        #all values
+        ind_low = ee_pulsed > 0
+        #no value
+        ind_high = ee_pulsed > 0
 
-    logger.info("We separate PF fit at %f keV" % e_turn)
+        if len(forced_gaussian_centroids) > 0:
+            center += [ [ee, (1-e_threshold)*ee, (1+e_threshold)*ee] for ee in forced_gaussian_centroids]
+            sigma += [[1+0*ee,0,3] for ee in forced_gaussian_centroids]
+            amplitude += [[0*ee,-2,2] for ee in forced_gaussian_centroids ]
 
-    if debug_plots:
-        axes[0].axvline(e_turn, 0, 1, color='cyan')
-        axes[1].axvline(e_turn, 0, 1, color='cyan')
 
-    ind_low = ee_pulsed <= e_turn
-    ind_high = ee_pulsed > e_turn
-    logger.info('We use %d points in the low-energy' % np.sum(ind_low))
 
     pulsed_fit_low = fit_pulsed_frac(ee_pulsed[ind_low],dee_pulsed[ind_low], 
                                         pulsed_frac[ind_low], dpulsed_frac[ind_low], n_gauss=1, 
-                                       center=[[6.5, 5.5, 7.5]],
-                                        sigma=[[0.2,0,1.0]],
-                                        amplitude=[[0,-1,1]],
-                                       degree_pol=poly_deg, plot_final=True, stem=stem+'low')
-
-    e_cyc = get_zeros(fake_en, smoothed_pulsed_frac_deriv, smoothed_pulsed_frac_deriv_2, fake_en>e_turn)
-    
-    if len(e_cyc)>1:
-        logger.info("Checking for too close Gaussians in %d elements" % len(e_cyc))
-        diff_e_cyc= ( 1- e_cyc[0:-1]/e_cyc[1:] )
-        ind_stay = diff_e_cyc > e_threshold
-#         print(e_cyc)
-#         print(diff_e_cyc)
-#         print(ind_stay)
-        e_cyc = [e_cyc[0]] + [ee for ee in e_cyc[1:][ind_stay]]
-        logger.info("Kept %d gaussians" % len(e_cyc))
-    
-    n_gauss = len(e_cyc)
-    if n_gauss >0:
-        center = [ [ee, (1-e_threshold)*ee, (1+e_threshold)*ee] for ee in e_cyc]
-        sigma=[[1+0*ee,0,3] for ee in e_cyc]
-        amplitude=[[0*ee,-2,2] for ee in e_cyc ]
-    else:
-        center = []
-        sigma = []
-        amplitude = []
-        
-    #print(center,sigma,amplitude)
-    logger.info('We use %d points in the high-energy' % np.sum(ind_high))
-    high_poly_deg = poly_deg
-    
-    if np.sum(ind_high) <= high_poly_deg + n_gauss * 3:
-        logger.info("Reducing polynomial order")
-        while (np.sum(ind_high) <= high_poly_deg + n_gauss * 3) and high_poly_deg >0:
-            
-            high_poly_deg -= 1
-        
-        if (np.sum(ind_high) <= high_poly_deg + n_gauss * 3):
-            n_gauss = 0
-            center = []
-            sigma = []
-            amplitude = []
-            high_poly_deg=1
-    
-    logger.info('We fit an order %d polynomial in the high-energy section' % high_poly_deg)
-    logger.info('We fit %d gaussians in the high-energy section' % n_gauss)
-    
-    pulsed_fit_high = fit_pulsed_frac(ee_pulsed[ind_high],dee_pulsed[ind_high], 
-                                        pulsed_frac[ind_high], dpulsed_frac[ind_high], n_gauss=n_gauss, 
                                        center=center,
                                         sigma=sigma,
                                         amplitude=amplitude,
-                                       degree_pol=high_poly_deg, plot_final=True, stem=stem+'high')
+                                       degree_pol=poly_deg, plot_final=True, stem=stem+'low')
+
+    if make_double_fit:
+
+        e_cyc = get_zeros(fake_en, smoothed_pulsed_frac_deriv, smoothed_pulsed_frac_deriv_2, fake_en>e_turn)
+        
+        if len(e_cyc)>1:
+            logger.info("Checking for too close Gaussians in %d elements" % len(e_cyc))
+            diff_e_cyc= ( 1- e_cyc[0:-1]/e_cyc[1:] )
+            ind_stay = diff_e_cyc > e_threshold
+    #         print(e_cyc)
+    #         print(diff_e_cyc)
+    #         print(ind_stay)
+            e_cyc = [e_cyc[0]] + [ee for ee in e_cyc[1:][ind_stay]]
+            logger.info("Kept %d gaussians" % len(e_cyc))
+        
+        n_gauss = len(e_cyc)
+        if n_gauss > 0 and max_n_high_lines > 0 and len(forced_gaussian_centroids) == 0:
+            center = [ [ee, (1-e_threshold)*ee, (1+e_threshold)*ee] for ee in e_cyc[0:max_n_high_lines]]
+            sigma=[[1+0*ee,0,3] for ee in e_cyc]
+            amplitude=[[0*ee,-2,2] for ee in e_cyc ]
+        elif len(forced_gaussian_centroids) > 0:
+            center = [ [ee, (1-e_threshold)*ee, (1+e_threshold)*ee] for ee in forced_gaussian_centroids]
+
+            if len(forced_gaussian_sigmas) == len(forced_gaussian_centroids):
+                sigma = [[ee,0,2*ee] for ee in forced_gaussian_sigmas ]
+            else:
+                sigma = [[1+0*ee,0,3] for ee in forced_gaussian_centroids]
+            
+            if len(forced_gaussian_amplitudes) == len(forced_gaussian_centroids):
+                amplitude = [[ee,-2*ee,4*ee] for ee in forced_gaussian_amplitudes ]
+            else:
+                amplitude = [[0*ee,-2,2] for ee in forced_gaussian_centroids ]
+            n_gauss = len(forced_gaussian_centroids)
+        else:
+            center = []
+            sigma = []
+            amplitude = []
+            n_gauss = 0
+            
+        #print(center,sigma,amplitude)
+        logger.info('We use %d points in the high-energy' % np.sum(ind_high))
+        high_poly_deg = poly_deg
+        
+        if np.sum(ind_high) <= high_poly_deg + n_gauss * 3:
+            logger.info("Reducing polynomial order")
+            while (np.sum(ind_high) <= high_poly_deg + n_gauss * 3) and high_poly_deg >0:
+                
+                high_poly_deg -= 1
+            
+            if (np.sum(ind_high) <= high_poly_deg + n_gauss * 3):
+                n_gauss = 0
+                center = []
+                sigma = []
+                amplitude = []
+                high_poly_deg=1
+        
+        logger.info('We fit an order %d polynomial in the high-energy section' % high_poly_deg)
+        logger.info('We fit %d gaussians in the high-energy section' % n_gauss)
+        
+        pulsed_fit_high = fit_pulsed_frac(ee_pulsed[ind_high],dee_pulsed[ind_high], 
+                                            pulsed_frac[ind_high], dpulsed_frac[ind_high], n_gauss=n_gauss, 
+                                        center=center,
+                                            sigma=sigma,
+                                            amplitude=amplitude,
+                                        degree_pol=high_poly_deg, plot_final=True, stem=stem+'high')
+    else:
+        pulsed_fit_high = None
 
     
     from matplotlib.pyplot import cm
     comps_low = pulsed_fit_low.eval_components(x=ee_pulsed[ind_low])
-    comps_high = pulsed_fit_high.eval_components(x=ee_pulsed[ind_high])
     bb_low = (pulsed_frac[ind_low] - pulsed_fit_low.best_fit) / dpulsed_frac[ind_low]
-    bb_high = (pulsed_frac[ind_high] - pulsed_fit_high.best_fit) / dpulsed_frac[ind_high]
+    if make_double_fit:
+        comps_high = pulsed_fit_high.eval_components(x=ee_pulsed[ind_high])
+        bb_high = (pulsed_frac[ind_high] - pulsed_fit_high.best_fit) / dpulsed_frac[ind_high]
 
     col = cm.viridis(np.linspace(0, 1, 6))
 
@@ -921,8 +1000,9 @@ def elaborate_pulsed_fraction(ee_pulsed, dee_pulsed, pulsed_frac, dpulsed_frac,
                 , fmt='.', color=col[4], label='data')
     ax1.plot(ee_pulsed[ind_low], pulsed_fit_low.best_fit, '-', label='best fit (low)', color=col[0])
     ax1.plot(ee_pulsed[ind_low], comps_low['poly_'], '--', label='Polynomial (low)', color=col[1])
-    ax1.plot(ee_pulsed[ind_high], pulsed_fit_high.best_fit, '-', label='best fit (high)', color=col[2])
-    ax1.plot(ee_pulsed[ind_high], comps_high['poly_'], '--', label='Polynomial (high)', color=col[3])
+    if make_double_fit:
+        ax1.plot(ee_pulsed[ind_high], pulsed_fit_high.best_fit, '-', label='best fit (high)', color=col[2])
+        ax1.plot(ee_pulsed[ind_high], comps_high['poly_'], '--', label='Polynomial (high)', color=col[3])
 
     ax1.legend(loc='upper left')
     ax1.set_xscale('log')
@@ -934,7 +1014,10 @@ def elaborate_pulsed_fraction(ee_pulsed, dee_pulsed, pulsed_frac, dpulsed_frac,
 
     ax2.set_ylabel('Residuals')
     ax1.set_ylim(-0.1, 1)
-    ax2.errorbar(ee_pulsed, np.concatenate([bb_low,bb_high]), xerr=dee_pulsed, yerr=1., fmt='.', color=col[2])
+    if make_double_fit:
+        ax2.errorbar(ee_pulsed, np.concatenate([bb_low,bb_high]), xerr=dee_pulsed, yerr=1., fmt='.', color=col[2])
+    else:
+        ax2.errorbar(ee_pulsed, bb_low, xerr=dee_pulsed, yerr=1., fmt='.', color=col[2])
     ax2.axhline(y=0, color='k', linestyle='--')
     if save_plot:
         plt.savefig(stem + 'global_pulsed_fitted.pdf')
@@ -1070,6 +1153,104 @@ def fit_pulsed_frac(en, den, pf, dpf, stem=None, degree_pol=4, n_gauss=0, center
     return out
 
 
+def explore_fit_mcmc(model_fitted, pars=[], hm_steps=7000, hm_walkers=2000, hm_jump=500, plot_corner=True, print_file=True,
+                     stem='posterior'):
+    '''
+    this function explore the posterior parameter space obtained with a previous fit
+    using mcmc
+    outputs: - corner plot
+             - Maximum Likelihood Estimation
+             - Error @ 1,2,3 sigma  
+
+
+    Parameters
+    ----------
+    model_fitted :  lmfit.model.ModelResult
+        output of the fit procedure obtained with
+        the function 'fit_pulsed_frac'
+    pars :  lmfit.parameter.Parameters
+        parameters of the fit (if not specified, we derive them from the results)
+    hm_steps : INTEGER
+        how many steps: number of steps for the MCMC, default = 7000
+    hm_walkers : INTEGER
+        how many walkers: number of walkers for MCMC, default = 2000
+    hm_jump : INTEGER
+        how many first step not to consider, default = 500
+    plot_corner : BOOLEAN
+        if the corner_plot has to be plotted
+
+    Returns
+    -------
+    explore: lmfit.minimizer.MinimizerResult (result of the minimization using emcee)
+    emcee_plot: corner plot of the posterior distribution
+
+
+    '''
+    import corner
+    if hm_steps <= hm_jump :
+        logger.error(" You almost did it!. However, the number of steps must be higher than the jumped ones:"
+                     "hm_steps >>> hm_jump")
+
+    if pars ==[]:
+        pars = model_fitted.params.copy()
+
+    explore = model_fitted.emcee(params=pars, steps=hm_steps, nwalkers=hm_walkers,
+                                is_weighted=True, burn=hm_jump)
+    if plot_corner:
+        emcee_plot = corner.corner(explore.flatchain, labels=explore.var_names, show_titles=True,
+                                   plot_datapoints=True)
+
+    highest_prob = np.argmax(explore.lnprob)
+    hp_loc = np.unravel_index(highest_prob, explore.lnprob.shape)
+    mle_soln = explore.chain[hp_loc]
+
+    pars2 = {}
+    for el in pars.keys():
+        if (pars[str(el)]).expr is None:
+            pars2[str(el)] = pars[str(el)]
+
+    for i, par in enumerate(pars2):
+        pars2[par].value = mle_soln[i]
+
+    fmt = '{:5s} {:11.5f} {:11.5f} {:11.5f}'.format
+
+    logger.info('\nMaximum Likelihood Estimation from emcee ')
+    logger.info('-------------------------------------------------')
+    logger.info('Parameter MLE Value Median Value Uncertainty')
+
+    for name, param in pars2.items():
+        logger.info(fmt(name, param.value, explore.params[name].value, explore.params[name].stderr))
+
+    logger.info('\Error estimates from emcee:')
+    logger.info('------------------------------------------------------')
+    logger.info('Parameter -3sigma -2sigma -1sigma median +1sigma +2sigma +3sigma')
+
+    if print_file:
+        o_f_emcee = open(stem + '.out', 'w')
+        o_f_emcee.write('# Maximum Likelihood Estimation from emcee \n#Parameter MLE Value Median Value Uncertainty \n')
+        for name, param in pars2.items():
+            o_f_emcee.write(fmt(name, param.value, explore.params[name].value, explore.params[name].stderr) + '\n')
+        o_f_emcee.write(
+            '#\Error estimates from emcee: \n#Parameter -3sigma -2sigma -1sigma median +1sigma +2sigma +3sigma \n')
+
+    for name in pars2.keys():
+        quantiles = np.percentile(explore.flatchain[name], [0.135, 2.275, 15.865, 50, 84.135, 97.725, 99.865])
+        median = quantiles[3]
+        err_m3 = quantiles[0] - median
+        err_m2 = quantiles[1] - median
+        err_m1 = quantiles[2] - median
+        err_p1 = quantiles[4] - median
+        err_p2 = quantiles[5] - median
+        err_p3 = quantiles[6] - median
+        fmt = '{:5s} {:8.4f} {:8.4f} {:8.4f} {:8.4f} {:8.4f} {:8.4f} {:8.4f}'.format
+        logger.info(fmt(name, err_m3, err_m2, err_m1, median, err_p1, err_p2, err_p3))
+        o_f_emcee.write(fmt(name, err_m3, err_m2, err_m1, median, err_p1, err_p2, err_p3) + '\n')
+
+    o_f_emcee.close()
+
+    return explore, emcee_plot
+
+
 def get_error_from_simul_rms(counts, counts_err, n_simul=100, n_harm=3):
 
     return get_error_from_simul(counts, counts_err, pulse_fraction_from_data_rms,
